# $Id: graphdefanglib.pl,v 1.3 2002/09/23 05:48:00 jpk Exp $
#
# GraphDefang -- a set of tools to create graphs of your mimedefang
#                spam and virus logs.
#
# Written by:    John Kirkland
#                jpk@bl.org
#
# Copyright (c) 2002, John Kirkland
#=============================================================================

use GD::Graph::linespoints;
use GD::Graph::bars;
use Date::Parse;
use Date::Format;
use Data::Dumper;
use strict;

# X and Y Graph Sizes in pixels

my $X_GRAPH_SIZE = 700;
my $Y_GRAPH_SIZE = 300;

# Number of hours, days, and months in the hourly, daily, and monthly charts, respectively.

my $NUM_HOURS_SUMMARY = 48;
my $NUM_DAYS_SUMMARY = 60;
my $NUM_MONTH_SUMMARY = 24;

# Output directory for png files that are created

my $OUTPUT_DIR = "/home/jpk/public_html/spam";

sub calculate
{
        my ($GraphSettings, $SpamLog, $VirusLog) = @_;

        my %VirusSummary = ();	# Accumulated Virus Values
        my %SpamSummary = ();	# Accumulated Spam Values
        my @GraphData = ();	# Array to send to graphing package

	# Set the graph title and filename according to the options chosen

	# Initialize the title
	$GraphSettings->{title} = "";
	$GraphSettings->{filename} = "";

	# Add "Top N" to the beginning of the Title if necessary
	if ($GraphSettings->{topN} > 1) { 
		$GraphSettings->{title} = "Top $GraphSettings->{topN} ";
	}

	# Set Spam and Virus Title
	if ($GraphSettings->{spam} && $GraphSettings->{virus}) {
		$GraphSettings->{title}		= $GraphSettings->{title} . "Spam & Virus Counts ";
		$GraphSettings->{filename} 	= "spamvirus";
	} elsif ($GraphSettings->{spam}) {
		$GraphSettings->{title}		= $GraphSettings->{title} . "Spam Counts ";
		$GraphSettings->{filename} 	= "spam";
	} elsif ($GraphSettings->{virus}) {
		$GraphSettings->{title} 	= $GraphSettings->{title} . "Virus Counts ";
		$GraphSettings->{filename} 	= "virus";
	} else {
		die ("Either GraphSettings{spam} or GraphSettings{virus} must be set to 1");
	}

	# Set Grouping Title
	if ($GraphSettings->{accumulate} eq 'summary') {
		$GraphSettings->{title} = $GraphSettings->{title} . "Total ";
	} elsif ($GraphSettings->{accumulate} eq 'virusname') {
		$GraphSettings->{title} = $GraphSettings->{title} . "by VirusName ";
	} elsif ($GraphSettings->{accumulate} eq 'sender') {
		$GraphSettings->{title} = $GraphSettings->{title} . "by Sender ";
	} elsif ($GraphSettings->{accumulate} eq 'recipient') {
		$GraphSettings->{title} = $GraphSettings->{title} . "by Recipient ";
	} elsif ($GraphSettings->{accumulate} eq 'subject') {
		$GraphSettings->{title} = $GraphSettings->{title} . "by Subject ";
	} else {
		die ("Invalid GraphSettings{accumulate} value");
	}

	$GraphSettings->{filename} = $GraphSettings->{filename} . "_$GraphSettings->{topN}$GraphSettings->{accumulate}_$GraphSettings->{type}";

	# The final portion of the title will be set in the GraphSettings{time} section below	

        if ($GraphSettings->{time} eq "hourly") {

                $GraphSettings->{DateAccumulator}	= "%Y%m%d%H"; #YYYYMMDDHH Format of the data accumulator string
                $GraphSettings->{NumValues}	= $NUM_HOURS_SUMMARY;	# Number of x-axis values on graph
                $GraphSettings->{NumSeconds} 	= 60*60;		# Incremental number of seconds represented by each x-axis value
                $GraphSettings->{DateFormatForXAxis} = "%h %d, %I%p";	# Format of date string on x-axis
                $GraphSettings->{x_label} 	= 'Hours';
                $GraphSettings->{y_label} 	= 'Counts per Hour';
		$GraphSettings->{title}		= $GraphSettings->{title} . "per Hour (last $GraphSettings->{NumValues} hours)";
                $GraphSettings->{filename} 	= "hourly_" . $GraphSettings->{filename};

		# Set Graph Title

        } elsif ($GraphSettings->{time} eq "daily") {

                $GraphSettings->{DateAccumulator}	="%Y%m%d";      #YYYYMMDD
                $GraphSettings->{NumValues}	= $NUM_DAYS_SUMMARY;
                $GraphSettings->{NumSeconds}	= 60*60*24;
                $GraphSettings->{DateFormatForXAxis} = "%h %d";
                $GraphSettings->{x_label} 	= 'Days';
                $GraphSettings->{y_label} 	= 'Counts per Day';
                $GraphSettings->{title} 		= $GraphSettings->{title} . "per Day (Last $GraphSettings->{NumValues} days)";
                $GraphSettings->{filename} 	= "daily_" . $GraphSettings->{filename};

        } elsif ($GraphSettings->{time} eq "monthly") {

                $GraphSettings->{DateAccumulator}	= "%Y%m";        #YYYYMM
                $GraphSettings->{NumValues} 	= $NUM_MONTH_SUMMARY;
                $GraphSettings->{NumSeconds} 	= 60*60*24*31;
                $GraphSettings->{DateFormatForXAxis} = "%h";
                $GraphSettings->{x_label} 	= 'Months';
                $GraphSettings->{y_label} 	= 'Counts per Month';
                $GraphSettings->{title}		= $GraphSettings->{title} . "per Month (Last $GraphSettings->{NumValues} months)";
                $GraphSettings->{filename} 	= "monthly_" . $GraphSettings->{filename};
        }

        # Calculate the date cutoff for our graph

	$GraphSettings->{CurrentTime} = time();
	$GraphSettings->{CutoffTime} = $GraphSettings->{CurrentTime} - ($GraphSettings->{NumSeconds}*($GraphSettings->{NumValues}-1));	

        #Accumulate the numbers of spam and/or viruses by the proper frequency

	if (($GraphSettings->{spam} == 1) and ($GraphSettings->{virus} == 1)) {
		$GraphSettings->{filename} = "$GraphSettings->{filename}_spamandvirus";
		@GraphData = accumulate_spamANDvirus($GraphSettings, \%$SpamLog, \%$VirusLog);
	} elsif ($GraphSettings->{spam} == 1) {
		$GraphSettings->{filename} = "$GraphSettings->{filename}_spam";
		@GraphData = accumulate_spamORvirus($GraphSettings, \%$SpamLog);
	} elsif ($GraphSettings->{virus} == 1) {
		$GraphSettings->{filename} = "$GraphSettings->{filename}_virus";
		@GraphData = accumulate_spamORvirus($GraphSettings, \%$VirusLog);
	} else {
		die ("Either Spam or Virus must be set to 1!");
	}

        # Draw Graph
	if ($GraphSettings->{type} eq 'line') {
        	draw_line_graph($GraphSettings,\@GraphData);
	} elsif ($GraphSettings->{type} eq 'stacked_bar') {
        	draw_stacked_bar_graph($GraphSettings,\@GraphData);
	} else {
		die ("Invalid GraphSettings{type} = $GraphSettings->{type}");
	}
}

sub accumulate_spamORvirus
{
        my ($GraphSettings, $DataLog) = @_;

        my %DataSummary = ();	# Accumulated Data Values
        my @GraphData = ();	# Array to send to graphing package
	my %Total = ();		# Accumulated totals for each accumulator (sender, recipient, etc...)

	# Set accumulator according to GraphSettings{accumulate}... choices
	# from log file are:  
	#       summary 		= 1
	#	hits or virusname	= 2
	#	sender			= 3
	#	recipient		= 4
	#	subject			= 5
	my $accumulator;

	if ($GraphSettings->{accumulate} eq 'summary') {
		$accumulator = 0;
	} elsif ($GraphSettings->{accumulate} eq 'hits') {
		$accumulator = 2;
	} elsif ($GraphSettings->{accumulate} eq 'virusname') {
		$accumulator = 2;
	} elsif ($GraphSettings->{accumulate} eq 'sender') {
		$accumulator = 3;
	} elsif ($GraphSettings->{accumulate} eq 'recipient') {
		$accumulator = 4;
	} elsif ($GraphSettings->{accumulate} eq 'subject') {
		$accumulator = 5;
	} else {
		die ("Bad GraphSettings{accumulate} value = $GraphSettings->{accumulate}");
	}

	if ($accumulator == 0) {		# Display counts, not grouped by anything

	        foreach my $i (keys %$DataLog) {
        	        my $date = time2str($GraphSettings->{DateAccumulator},$DataLog->{$i}->[0]);
                	$DataSummary{$date}++;
			$Total{summary}++;
	        }

        	# Create data array for graphing

		for (my $time=$GraphSettings->{CutoffTime};
                        $time<=$GraphSettings->{CurrentTime};
                        $time += $GraphSettings->{NumSeconds}
                    ) {
        	        my $date = time2str($GraphSettings->{DateAccumulator},$time);
                	my $datestring = time2str($GraphSettings->{DateFormatForXAxis},$time);

	                push @{$GraphData[0]}, $datestring;
	                push @{$GraphData[1]}, $DataSummary{$date};
		}

		if ($GraphSettings->{spam} == 1) {
			push ( @{$GraphSettings->{legend}}, ("Spam Count - Total: $Total{summary}") );
		} elsif ($GraphSettings->{virus} == 1) {
			push ( @{$GraphSettings->{legend}}, ("Virus Count - Total: $Total{summary}") );
		} else {
			die ("uhh... how did that happen?!");
		}

	} else {				# Display data grouped by accumulator value
	
        	foreach my $i (keys %$DataLog) {
			# Don't count data that is outside of our range.
			if ($DataLog->{$i}->[0] > $GraphSettings->{CutoffTime}) {
                		my $date = time2str($GraphSettings->{DateAccumulator},$DataLog->{$i}->[0]);
	                	my $DataName = lc($DataLog->{$i}->[$accumulator]); #lower case all of the comparisons
        	        	$DataSummary{$date}{$DataName}++;
				$Total{$DataName}++;
			}
	        }
	
        	# Get unique data list

        	my @DataList = ();
		
		# Add totals to the legend.
		my @Legend = ();

		# Sort the total array, and keep only the top N of the $DataSummary hash
		foreach my $DataName (sort { $Total{$b} <=> $Total{$a} } keys %Total) {
			push @DataList, $DataName;
			push @Legend, "$DataName, Total: $Total{$DataName}";
		}
		
		# Take care of any topN settings.
		if ($GraphSettings->{topN} > 0) {
			$#DataList = $GraphSettings->{topN} - 1;
		}
		# Create data Array

	        for (my $time=$GraphSettings->{CutoffTime}; 
			$time<=$GraphSettings->{CurrentTime}; 
			$time += $GraphSettings->{NumSeconds}
		    ) {

	                my $date = time2str($GraphSettings->{DateAccumulator},$time);
	                my $datestring = time2str($GraphSettings->{DateFormatForXAxis},$time);

	                my $i=0;
	                push @{$GraphData[$i]}, $datestring;
	                foreach my $DataName (@DataList) {
	                        $i++;
	                        push @{$GraphData[$i]}, $DataSummary{$date}{$DataName};
	                }
	        }
		@{$GraphSettings->{legend}} = @Legend;
	}
	return @GraphData;
}

sub accumulate_spamANDvirus
{

        my ($GraphSettings, $SpamLog, $VirusLog) = @_;

        my %VirusSummary = ();	# Accumulated Virus Values
        my %SpamSummary = ();	# Accumulated Spam Values
        my @GraphData = ();	# Array to send to graphing package
	my %Total = ();		# Accumulated totals for each accumulator (sender, recipient, etc...)

        # Set accumulator according to GraphSettings{accumulate}... choices
        # from log file are:
        #       summary                 = 0
        #       hits or virusname       = 2
        #       sender                  = 3
        #       recipient               = 4
        #       subject                 = 5
        my $accumulator;

        if ($GraphSettings->{accumulate} eq 'summary') {
                $accumulator = 0;
#        } elsif ($GraphSettings->{accumulate} eq 'hits') {	# It doesn't make sense to look at spamANDviruses by these
#                $accumulator = 2;				
#        } elsif ($GraphSettings->{accumulate} eq 'virusname') {
#                $accumulator = 2;
        } elsif ($GraphSettings->{accumulate} eq 'sender') {
                $accumulator = 3;
        } elsif ($GraphSettings->{accumulate} eq 'recipient') {
                $accumulator = 4;
        } elsif ($GraphSettings->{accumulate} eq 'subject') {
                $accumulator = 5;
        } else {
                die ("Invalid GraphSettings{accumulate} value = $GraphSettings->{accumulate}");
        }

        if ($accumulator == 0) {         # We are displaying counts, not grouped by anything


	        # Draw Overall Summary Graph

        	foreach my $i (keys %$SpamLog) {
               	 	my $date = time2str($GraphSettings->{DateAccumulator},$SpamLog->{$i}->[0]);
	                $SpamSummary{$date}++;
			$Total{spam}++;
	        }

	        foreach my $i (keys %$VirusLog) {
	                my $date = time2str($GraphSettings->{DateAccumulator},$VirusLog->{$i}->[0]);
	                $VirusSummary{$date}++;
			$Total{virus}++;
	        }

	        # Create data array for graphing

	        for (my $time=$GraphSettings->{CutoffTime}; 
			$time<=$GraphSettings->{CurrentTime}; 
			$time += $GraphSettings->{NumSeconds}
		    ) {

	                my $date = time2str($GraphSettings->{DateAccumulator},$time);
	                my $datestring = time2str($GraphSettings->{DateFormatForXAxis},$time);

	                push @{$GraphData[0]}, $datestring;
	                push @{$GraphData[1]}, $SpamSummary{$date};
	                push @{$GraphData[2]}, $VirusSummary{$date};

        	}

		# Graph Legend
		push ( @{$GraphSettings->{legend}}, ("Spam Count - Total: $Total{spam}", "Virus Count - Total: $Total{virus}") );
	} else {
		# Here we accumulate data for both SPAM and VIRUSES for sender, recipient or subject data.
		# I have left out hits and virusname because they don't seem to make sense in this context.
		# For topN, we accumulate a total of both spam and viruses together and take the topN offenders.

		my %DataSummary = ();
		my %SpamTotal = ();
		my %VirusTotal = ();

		# Accumulate spam data
                foreach my $i (keys %$SpamLog) {
                        # Don't count data that is outside of our range.
                        if ($SpamLog->{$i}->[0] > $GraphSettings->{CutoffTime}) {
                                my $date = time2str($GraphSettings->{DateAccumulator},$SpamLog->{$i}->[0]);
                                my $DataName = lc($SpamLog->{$i}->[$accumulator]); #lower case all of the comparisons
                                $DataSummary{$date}{$DataName}++;
                                $Total{$DataName}++;	# %Total for calculating topN
				$SpamTotal{$DataName}++;  # Spam only total for legend
                        }
                }

		# Accumulate virus data
                foreach my $i (keys %$VirusLog) {
                        # Don't count data that is outside of our range.
                        if ($VirusLog->{$i}->[0] > $GraphSettings->{CutoffTime}) {
                                my $date = time2str($GraphSettings->{DateAccumulator},$VirusLog->{$i}->[0]);
                                my $DataName = lc($VirusLog->{$i}->[$accumulator]); #lower case all of the comparisons
                                $DataSummary{$date}{$DataName}++;
                                $Total{$DataName}++;	# %Total for calculating topN
				$VirusTotal{$DataName}++;  # Virus only total for legend
                        }
                }

                # Get unique data list

                my @DataList = ();

                # Add totals to the legend.
		my $LegendString;
                my @Legend = ();

                # Sort the total array, and keep only the top N 
                foreach my $DataName (sort { $Total{$b} <=> $Total{$a} } keys %Total) {
                        push @DataList, $DataName;
			if ( exists($SpamTotal{$DataName}) ) {
				$LegendString = "$DataName, Spam: $SpamTotal{$DataName}";
				if ( exists ($VirusTotal{$DataName}) ) {
					$LegendString = $LegendString . ", Virus: $VirusTotal{$DataName}";
				}
			}
			elsif ( exists($VirusTotal{$DataName}) ) {
				$LegendString = "Virus: $VirusTotal{$DataName}";
			}
				
                        push @Legend, $LegendString;
                }

                # Take care of any topN settings.
                if ($GraphSettings->{topN} > 0) {
                        $#DataList = $GraphSettings->{topN} - 1;
			$#Legend = $GraphSettings->{topN} - 1;
                }
                # Create data Array

                for (my $time=$GraphSettings->{CutoffTime};
                        $time<=$GraphSettings->{CurrentTime};
                        $time += $GraphSettings->{NumSeconds}
                    ) {

                        my $date = time2str($GraphSettings->{DateAccumulator},$time);
                        my $datestring = time2str($GraphSettings->{DateFormatForXAxis},$time);

                        my $i=0;
                        push @{$GraphData[$i]}, $datestring;
                        foreach my $DataName (@DataList) {
                                $i++;
                                push @{$GraphData[$i]}, $DataSummary{$date}{$DataName};
                        }
                }
                @{$GraphSettings->{legend}} = @Legend;
	}
	return @GraphData;
}

sub draw_line_graph
{
	my ($Settings, $Data) = @_;

	my $my_graph = new GD::Graph::linespoints($X_GRAPH_SIZE, $Y_GRAPH_SIZE);

	$my_graph->set(
        	x_label 		=> $Settings->{x_label},
	        y_label 		=> $Settings->{y_label},
	        title 			=> $Settings->{title}, 
	        x_labels_vertical 	=> 1,
	        x_label_position 	=> 1/2,

		bgclr			=> 'white',
		fgclr			=> 'gray',
		boxclr			=> 'lgray',

		y_tick_number		=> 10,
		y_label_skip		=> 2,
		long_ticks 		=> 1,
	        marker_size 		=> 1,
	        skip_undef 		=> 1,
	        line_width 		=> 1,
	        transparent 		=> 0,
	);

#print Dumper($my_graph);
	
	$my_graph->set_legend( @{$Settings->{legend}} );
	$my_graph->plot($Data);
	save_chart($my_graph, "$OUTPUT_DIR/$Settings->{filename}");
}

sub draw_stacked_bar_graph
{
        my ($Settings, $Data) = @_;

        my $my_graph = new GD::Graph::bars($X_GRAPH_SIZE, $Y_GRAPH_SIZE);

        $my_graph->set(
                x_label 		=> $Settings->{x_label},
                y_label 		=> $Settings->{y_label},
                title 			=> $Settings->{title},
                x_labels_vertical 	=> 1,
                x_label_position 	=> 1/2,
		
		bgclr			=> 'white',
		fgclr			=> 'gray',
		boxclr			=> 'lgray',

		y_tick_number		=> 10,
		y_label_skip		=> 2,
		long_ticks 		=> 1,
                cumulate        	=> 1,
                transparent     	=> 0,
		correct_width		=> 0,
        );

        $my_graph->set_legend( @{$Settings->{legend}} );
        $my_graph->plot($Data);
        save_chart($my_graph, "$OUTPUT_DIR/$Settings->{filename}");
}

sub read_data
{
	my $fn = shift;
	my %d = ();
	my $i = 0;

	open(ZZZ, $fn) || return ();

	while (<ZZZ>)
	{
		chomp;
		my @row = split /,/, $_, 5; #split line on commas
		my @subrow = split / /, $row[0]; #split up date and other items at the beginning of the line

		my $datestring = "$subrow[0] $subrow[1] $subrow[2]";
		my $date = str2time($datestring);
		my $msgID = $subrow[5];
		my $HitsOrVirusName = $row[1];
		my $sender = $row[2];
		my $recipient = $row[3];
		my $subject = $row[4];

		push( @{$d{$i}}, ($date,$msgID,$HitsOrVirusName,$sender,$recipient,$subject));

		$i++;
	}

	close (ZZZ);

	return %d;
}

sub save_chart
{
	my $chart = shift or die "Need a chart!";
	my $name = shift or die "Need a name!";
	local(*OUT);

	my $ext = $chart->export_format;

	open(OUT, ">$name.png") or 
		die "Cannot open $name.$ext for write: $!";
	binmode OUT;
	print OUT $chart->gd->png;
	close OUT;
}
1;
